IMPLEMENTATION MODULE Client;

FROM SYSTEM IMPORT BYTE, ADR;
IMPORT Str, Types, Lib2, Lib3, SO, Crc, MD5, Tmrs, Tmr, UFTP, Config, SrvIO, File, Cfg, Log;

CONST
  mid = Cfg.midClient;

  dtRnd  = 20;          (* ms, round-trip time (estimated) *)
  toSes  = 50 * dtRnd;  (* ms, session request period *)
  toDat  =  5 * dtRnd;  (* ms, data request period *)
  fglMax = UFTP.lenMaxFgm;
  wndMax = 64;

  cntMax = 16;  (* general probe count *)
  cntDtr =  8;  (* round trip time probe count *)
  cntFpu =  8;  (* fragments per update probe count *)

TYPE
  TState = ( stOpen, stData, stClose );

  TSesDesc = RECORD
    t: Tmr.TTicks;
    f: UFTP.TFgmLen;
    l: UFTP.TFileLen;
    d: MD5.TDgstDesc;
    r: UFTP.TReference;
    n: UFTP.TFgmIdx;  (* fgm cnt *)
    e: UFTP.TFgmLen;  (* last fgm *)
  END;

  TWndLen = UFTP.TFgmIdx[ 0 .. wndMax ];
  TWndIdx = UFTP.TFgmIdx[ 0 .. wndMax - 1 ];

  TWndDesc = RECORD
    i: UFTP.TFgmIdx;  (* lower window boundary *)
    j: UFTP.TFgmIdx;  (* upper window boundary + 1 *)
    w: TWndLen;       (* max window size *)
    (* rx infos *)
    d: SET OF TWndIdx;  (* done fragments, exact *)
    n: TWndLen;         (* done fragments, count *)
    (* tx infos *)
    t: ARRAY TWndIdx OF Tmr.TTicks;  (* request times *)
  END;
  (* 0 <= i <= j <= ses.n *)
  (* j - i <= w <= wndMax *)

  TAvgVal = Types.TLngCard;
  TPrbCnt = Types.TShtCard[ 0 .. cntMax ];
  TPrbIdx = Types.TShtCard[ 0 .. cntMax - 1 ];

  TAvgDesc = RECORD
    v: TAvgVal;
    i: TPrbIdx;   (* write pos *)
    n: TPrbCnt;
    m: TPrbCnt;   (* max *)
    a: ARRAY TPrbIdx OF TAvgVal;
  END;

  TDataDesc = RECORD
    st : TState;
    ers: TErrCnts;
    ses: TSesDesc;
    wnd: TWndDesc;
    dtr: TAvgDesc;
    fpu: TAvgDesc;
  END;

VAR
  data: TDataDesc;
  tmr: Tmrs.TTimer;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlMsg, s, Config.data.vrb );
END Message;

PROCEDURE Error( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlErr, s, Config.data.vrb );
  HALT;  (* solve return code problem before removing this line *)
END Error;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlFat, s, TRUE );
  HALT;
END Fatal;


PROCEDURE ShowErrCnts();
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
  e: TErr;
BEGIN
  SO.Init( b, s );
  SO.WrStr( b, 'ShowErrCnts' );
  FOR e := MIN(TErr) TO MAX(TErr) DO
    SO.WrChar( b, ' ' );
    SO.WrLngHex( b, data.ers[e] MOD 1000H, 3 );
  END;
  Message( s );
END ShowErrCnts;


PROCEDURE AvgInit( VAR a: TAvgDesc; x: TAvgVal; m: TPrbCnt );
BEGIN
  IF m = 0 THEN Fatal( 'AvgInit: Invalid max cnt' ) END;
  IF x > MAX( TAvgVal ) DIV ( 2 * m ) THEN Fatal( 'AvgInit: Invalid val' ) END;
  Lib2.Fill( a, 0 );
  a.v := x;
  a.i := 1;
  a.n := 1;
  a.m := m;
  a.a[0] := x;
END AvgInit;

PROCEDURE AvgUpd( VAR a: TAvgDesc; x: TAvgVal );
VAR
  i: TPrbCnt;
  v: TAvgVal;
BEGIN
  IF x > MAX( TAvgVal ) DIV ( 2 * a.m ) THEN Fatal( 'AvgUpd: Invalid val' ) END;
  i := a.i;
  a.a[ i ] := x;
  a.i := ( i + 1 ) MOD a.m;
  IF a.n < a.m THEN INC( a.n ) END;
  v := a.a[ 0 ];
  FOR i := 1 TO a.n - 1 DO INC( v, a.a[i] ) END;
  a.v := v DIV a.n;
END AvgUpd;


PROCEDURE WrWnd( VAR b: SO.TBuffer );
VAR
  i: UFTP.TFgmIdx;
  k: TWndIdx;
BEGIN
  SO.WrStr( b, ' wnd ' );
  SO.WrLngCard( b, data.wnd.i, 1 );
  SO.WrChar( b, ' ' );
  SO.WrLngCard( b, data.wnd.j, 1 );
  SO.WrChar( b, ' ' );
  SO.WrLngCard( b, data.wnd.j - data.wnd.i, 1 );
  SO.WrChar( b, ' ' );
  SO.WrLngCard( b, data.wnd.w, 1 );
  SO.WrChar( b, ' ' );
  SO.WrLngCard( b, data.wnd.n, 1 );
  (*
  SO.WrStr( b, ' don ' );
  i := data.wnd.i;
  WHILE i < data.wnd.j DO
    k := i MOD wndMax;
    IF k IN data.wnd.d THEN
      SO.WrLngCard( b, i, 1 );
      SO.WrChar( b, ' ' );
    END;
    INC( i );
  END;
  *)
END WrWnd;


PROCEDURE IncErrCnt( e: TErr );
BEGIN
  Lib2.IncL( data.ers[e], 1 );
  Lib2.IncL( data.ers[eNull], 1 );
END IncErrCnt;

PROCEDURE OnRx( dat-: ARRAY OF BYTE; len: CARDINAL );
VAR
  tel: POINTER TO UFTP.TTelDesc;

  PROCEDURE RdRp();
  CONST
    lenMin = SIZE( UFTP.TRdRpDesc );
    lenMax = lenMin;
    dtr = dtRnd;
  VAR
    f: UFTP.TFgmLen;
    l: UFTP.TFileLen;
    w: UFTP.TFgmIdx;
    t, dt: Tmr.TTicks;

    PROCEDURE Show();
    CONST
      ena = TRUE;
    VAR
      b: SO.TBuffer;
      s: ARRAY [0..127] OF CHAR;
      m: MD5.TDgstStr;
    BEGIN
      IF ~ ena THEN RETURN END;
      MD5.MkStr( m, data.ses.d );
      SO.Init( b, s );
      SO.WrStr( b, 'RdRp: fgl ' );
      SO.WrLngCard( b, data.ses.f, 1 );
      SO.WrStr( b, ' len ' );
      SO.WrLngCard( b, data.ses.l, 1 );
      SO.WrStr( b, ' md5 ' );
      SO.WrStr( b, m );
      SO.WrStr( b, ' ref ' );
      SO.WrLngCard( b, data.ses.r, 1 );
      SO.WrStr( b, ' cnt ' );
      SO.WrLngCard( b, data.ses.n, 1 );
      SO.WrStr( b, ' lst ' );
      SO.WrLngCard( b, data.ses.e, 1 );
      SO.WrStr( b, ' dt ' );
      SO.WrLngCard( b, dt, 1 );
      SO.WrStr( b, ' w ' );
      SO.WrLngCard( b, w, 1 );
      Message( s );
    END Show;

  BEGIN
    IF data.st # stOpen THEN RETURN END;
    IF ( len < lenMin ) OR ( len > lenMax ) THEN
      IncErrCnt( eRdRpLen );
      RETURN;
    END;
    CASE tel^.rrp.e OF
    | UFTP.errNull: (* ok *)
    | UFTP.errRID : Fatal( 'Resource not found.' );
    | UFTP.errRes : Fatal( 'Server full.' );
    ELSE
      Fatal( 'Unknown error.' );
    END;
    t := Tmr.Ticks();
    f := tel^.rrp.f;
    l := tel^.rrp.p.l;
    IF ( f = 0 ) OR ( f > fglMax ) THEN
      IncErrCnt( eRdRpFgl );
      RETURN;
    END;
    dt := Tmr.Diff( data.ses.t, t );
    AvgInit( data.dtr, dt, cntDtr );
    data.ses.t := t;
    data.ses.f := f;
    data.ses.l := l;
    data.ses.d := tel^.rrp.p.d;
    data.ses.r := tel^.rrp.r;
    data.ses.n := l DIV f + ORD( l MOD f > 0 );
    data.ses.e := l MOD f;
    Lib2.Fill( data.wnd, 0 );
    (* data.wnd.w should be fgf * dtr (fragment frequency times round-trip time) *)
    (* fgf = fpu / dtu *)
    w := Config.data.fpu * dtr DIV Config.data.dtu;
    IF w > wndMax THEN w := wndMax END;
    data.wnd.w := w;
    File.Open( 'out.bin', f, l, data.ses.d );
    IF data.ses.n > 0 THEN data.st := stData ELSE data.st := stClose END;
    Show();
  END RdRp;

  PROCEDURE DatRp();
  CONST
    lenMin = UFTP.lenMinDatRpDesc;
    lenMax = UFTP.lenMaxDatRpDesc;
  VAR
    i, m: UFTP.TFgmIdx;
    k: TWndIdx;
    dt: Tmr.TTicks;

    PROCEDURE Show();
    CONST
      ena = TRUE;
    VAR
      b: SO.TBuffer;
      s: ARRAY [0..511] OF CHAR;
    BEGIN
      IF ~ ena THEN RETURN END;
      SO.Init( b, s );
      SO.WrStr( b, 'DatRp: idx ' );
      SO.WrLngCard( b, i, 1 );
      SO.WrStr( b, ' len ' );
      SO.WrLngCard( b, len, 1 );
      SO.WrStr( b, ' dt ' );
      SO.WrLngCard( b, dt, 1 );
      SO.WrStr( b, ' dtr ' );
      SO.WrLngCard( b, data.dtr.v, 1 );
      WrWnd( b );
      Message( s );
    END Show;

  BEGIN
    IF data.st # stData THEN RETURN END;
    IF ( len < lenMin ) OR ( len > lenMax ) THEN
      IncErrCnt( eDatRpLen );
      RETURN;
    END;
    DEC( len, SIZE( UFTP.TDatHdrDesc ) );
    i := tel^.drp.h.i;
    IF ( i < data.wnd.i ) OR ( i >= data.wnd.j ) THEN
      IncErrCnt( eDatRpWnd );  (* not really an error, just for debugging *)
      RETURN;
    END;
    m := data.ses.n - 1;  (* max fgm idx *)
    IF ( i < m ) & ( len # data.ses.f ) OR
       ( i = m ) & ( len # data.ses.e ) THEN
      IncErrCnt( eDatRpFgl );
      RETURN;
    END;
    k := i MOD wndMax;
    IF k IN data.wnd.d THEN
      IncErrCnt( eDatRpDup );  (* not really an error, just for debugging *)
      dt := 0;
    ELSE
      File.Write( i, tel^.drp.d, len );
      INCL( data.wnd.d, k );
      INC( data.wnd.n );
      dt := Tmr.Diff( data.wnd.t[k], Tmr.Ticks() );
      AvgUpd( data.dtr, dt );
    END;
    Show();
  END DatRp;

CONST
  lenMin = UFTP.lenFrmOvr;
  lenMax = UFTP.lenMaxTel;
BEGIN
  IF ( len < lenMin ) OR ( len > lenMax ) THEN IncErrCnt( eLen ); RETURN END;
  IF ~ Crc.Ok( dat, len ) THEN IncErrCnt( eCrc ); RETURN END;
  tel := ADR( dat );
  DEC( len, lenMin );
  IF tel^.hdr.m # UFTP.mag THEN IncErrCnt( eMag ); RETURN END;
  IF tel^.hdr.v # UFTP.ver THEN IncErrCnt( eVer ); RETURN END;
  CASE tel^.hdr.c OF
  | UFTP.cmdRdRp : RdRp();
  | UFTP.cmdDatRp: DatRp();
  ELSE
    IncErrCnt( eCmd );
  END;
END OnRx;


PROCEDURE DoOpen();
VAR
  r: RECORD h: UFTP.THdrDesc; b: UFTP.TRdRqDesc; t: UFTP.TTrlDesc END;

  PROCEDURE ShowTel();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'RdRq: fgl ' );
    SO.WrCard( b, r.b.f, 1 );
    SO.WrStr( b, " rid '" );
    SO.WrStr( b, r.b.r );
    SO.WrStr( b, "'" );
    Message( s );
  END ShowTel;

VAR
  t: Tmr.TTicks;
BEGIN
  t := Tmr.Ticks();
  IF Tmr.Diff( data.ses.t, t ) < toSes THEN RETURN END;
  data.ses.t := t;
  AvgInit( data.fpu, 0, cntFpu );
  Lib2.Fill( r, 0 );
  r.h.m := UFTP.mag;
  r.h.v := UFTP.ver;
  r.h.c := UFTP.cmdRdRq;
  r.b.f := fglMax;
  Str.Copy( r.b.r, Config.data.rnm );
  r.t := Crc.Swap( Crc.UpdL( Crc.crcInit, r, SIZE(r) - SIZE(r.t) ) );
  SrvIO.Send( r, SIZE( r ) );
  ShowTel();
END DoOpen;

PROCEDURE DoData();
VAR
  tel: RECORD h: UFTP.THdrDesc; b: UFTP.TDatRqDesc; t: UFTP.TTrlDesc END;

  PROCEDURE ShowTel();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'DatRq: idx ' );
    SO.WrLngCard( b, tel.b.h.i, 1 );
    SO.WrStr( b, ' set ' );
    SO.WrLngHex( b, LONGCARD( tel.b.s ), 8 );
    Message( s );
  END ShowTel;

  PROCEDURE Flush();
  BEGIN
    IF ( tel.h.m = 0 ) OR ( tel.b.s = UFTP.TFgmSet{} ) THEN RETURN END;
    tel.t := Crc.Swap( Crc.UpdL( Crc.crcInit, tel, SIZE(tel) - SIZE(tel.t) ) );
    SrvIO.Send( tel, SIZE( tel ) );
    ShowTel();
  END Flush;

  PROCEDURE Request( i: UFTP.TFgmIdx );
  BEGIN
    IF tel.h.m = 0 THEN
      tel.h.m := UFTP.mag;
      tel.h.v := UFTP.ver;
      tel.h.c := UFTP.cmdDatRq;
      tel.b.h.r := data.ses.r;
      tel.b.s := UFTP.TFgmSet{};
    END;
    IF tel.b.s = UFTP.TFgmSet{} THEN tel.b.h.i := i END;
    IF i - tel.b.h.i > MAX( UFTP.TRelFgmIdx ) THEN
      Flush();
      tel.b.h.i := i;
      tel.b.s := UFTP.TFgmSet{};
    END;
    INCL( tel.b.s, i - tel.b.h.i );
  END Request;

  PROCEDURE ShowWnd( n: CARDINAL );
  CONST
    ena = FALSE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..511] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'DoData' );
    SO.WrCard( b, n, 1 );
    SO.WrStr( b, ': fpu ' );
    SO.WrLngCard( b, data.fpu.v, 1 );
    WrWnd( b );
    Message( s );
  END ShowWnd;

  PROCEDURE ShowTO( i: UFTP.TFgmIdx );
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..511] OF CHAR;
  BEGIN
    IncErrCnt( eTimeOut );
    SO.Init( b, s );
    SO.WrStr( b, 'DoData.TO: idx ' );
    SO.WrLngCard( b, i, 1 );
    Message( s );
  END ShowTO;

VAR
  t: Tmr.TTicks;
  i, j: UFTP.TFgmIdx;
  k: TWndIdx;
  n: CARDINAL;
BEGIN
  tel.h.m := 0;
  t := Tmr.Ticks();
  AvgUpd( data.fpu, data.wnd.n );
  ShowWnd( 1 );
  (* clear left = progress wnd.i *)
  i := data.wnd.i;
  k := i MOD wndMax;
  WHILE ( i < data.wnd.j ) & ( k IN data.wnd.d ) DO
    EXCL( data.wnd.d, k );
    DEC( data.wnd.n );
    INC( i );
    k := i MOD wndMax;
  END;
  data.wnd.i := i;
  ShowWnd( 2 );
  (* re-request *)
  n := Config.data.fpu;
  WHILE ( i < data.wnd.j ) & ( n > 0 ) DO
    k := i MOD wndMax;
    IF ~ ( k IN data.wnd.d ) & ( Tmr.Diff( data.wnd.t[k], t ) >= toDat ) THEN
      ShowTO( i );
      Request( i );
      data.wnd.t[ k ] := t;
      DEC( n );
    END;
    INC( i );
  END;
  (* adjust wnd.w *)
  (* request new = progress wnd.j *)
  j := data.wnd.i + data.wnd.w;
  IF j > data.ses.n THEN j := data.ses.n END;
  i := data.wnd.j;
  WHILE ( n > 0 ) & ( i < j ) DO
    k := i MOD wndMax;
    Request( i );
    data.wnd.t[ k ] := t;
    DEC( n );
    INC( i );
  END;
  data.wnd.j := i;
  Flush();
  ShowWnd( 3 );
  (* close if done *)
  IF data.wnd.i >= data.ses.n THEN data.st := stClose END;
END DoData;

PROCEDURE DoClose();
VAR
  r: RECORD h: UFTP.THdrDesc; b: UFTP.TCloseDesc; t: UFTP.TTrlDesc END;

  PROCEDURE ShowTel();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Close: ref ' );
    SO.WrCard( b, r.b.r, 1 );
    Message( s );
  END ShowTel;

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
    dt: Tmr.TTicks;
  BEGIN
    IF ~ ena THEN RETURN END;
    dt := Tmr.Ticks() - data.ses.t;
    SO.Init( b, s );
    SO.WrStr( b, 'DoClose: ' );
    SO.WrLngCard( b, data.ses.l, 1 );
    SO.WrStr( b, ' bytes / ' );
    SO.WrLngCard( b, dt, 1 );
    SO.WrStr( b, ' ms = ' );
    SO.WrLngCard( b, data.ses.l DIV dt, 1 );
    SO.WrStr( b, ' bytes/ms' );
    Message( s );
  END Show;

BEGIN
  File.Close();

  Lib2.Fill( r, 0 );
  r.h.m := UFTP.mag;
  r.h.v := UFTP.ver;
  r.h.c := UFTP.cmdClose;
  r.b.r := data.ses.r;
  r.t := Crc.Swap( Crc.UpdL( Crc.crcInit, r, SIZE(r) - SIZE(r.t) ) );
  SrvIO.Send( r, SIZE( r ) );
  ShowTel();

  Show();
  Lib3.SetReturnCode( 0 );
  HALT;
END DoClose;

PROCEDURE OnTmr( k: Tmrs.TTicks; r: Tmrs.TRef );
BEGIN
  CASE data.st OF
  | stOpen : DoOpen();
  | stData : DoData();
  | stClose: DoClose();
  ELSE
    Fatal( 'OnTmr: Invalid state' );
  END;
END OnTmr;


PROCEDURE Run();
BEGIN
  data.ses.t := Tmr.Sub( Tmr.Ticks(), toSes );  (* force initial open timeout *)
  SrvIO.Execute();
END Run;

PROCEDURE GetErrCnts( VAR e: TErrCnts );
BEGIN
  e := data.ers;
END GetErrCnts;


VAR
  oldTerm: PROC;

PROCEDURE DoneModule();
BEGIN
  ShowErrCnts();
  Message( 'DoneModule' );
  oldTerm();
END DoneModule;

PROCEDURE InitModule();

  PROCEDURE Show();
  CONST
    ena = TRUE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [1..64] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'InitModule UFTP.ver = ' );
    SO.WrCard( b, UFTP.ver, 1 );
    Message( s );
  END Show;

BEGIN
  Log.Register( mid, 'Client' );
  Lib2.Fill( data, 0 );

  tmr := Tmrs.New();
  Tmrs.Set( tmr, 1, Config.data.dtu );
  Tmrs.SetHnd( tmr, OnTmr, 0 );

  SrvIO.Install( OnRx );

  Lib3.Terminate( DoneModule, oldTerm );

  Show();
END InitModule;

BEGIN
  InitModule();
END Client.
