(* UFTP senders *)
(* solyga@gmx.de 2022-01-07 *)
IMPLEMENTATION MODULE Sndrs;

FROM SYSTEM IMPORT BYTE, ADR;
IMPORT Lib2, Lib3, FIO, FIOStd, Str;
IMPORT UFTP, MD5, Cfg, SO, Log, Config;

CONST
  period = 1;  (* ms *)
  mid = Cfg.midSndrs;

TYPE
  TCnt = [ 0 .. cntMax ];
  TIdx = TCnt;
  TVldIdx = TIdx[ 1 .. MAX( TIdx ) ];

  TPthStr = ARRAY [ 1 .. 2 * SIZE( UFTP.TResID ) ] OF CHAR;

TYPE
  TFile = POINTER TO TFileDesc;
  TFileDesc = RECORD
    i: TIdx;           (* 0 <==> free *)
    b: FIO.File;
    r: UFTP.TResID;
    t: UFTP.TPosDesc;  (* target *)
  END;

  TFilesDesc = RECORD
    n: TCnt;
    a: ARRAY TVldIdx OF TFileDesc;
  END;

CONST
  cntMaxReq = 64;

TYPE
  TReqsDesc = RECORD
    n,
    r,
    w: CARDINAL;
    a: ARRAY [ 0 .. cntMaxReq - 1 ] OF UFTP.TFgmIdx;
    h: UFTP.TFgmIdx;
  END;

TYPE
  TSndr = POINTER TO TSndrDesc;
  TSndrDesc = RECORD
    i: TIdx;           (* 0 <==> free *)
    l: UFTP.TFgmLen;
    r: TReqsDesc;
    f: TFile;
  END;

  TSndrsDesc = RECORD
    n: TCnt;
    a: ARRAY TVldIdx OF TSndrDesc;
  END;

VAR
  fil: TFilesDesc;
  snd: TSndrsDesc;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlMsg, s, Config.data.vrb );
END Message;

PROCEDURE Error( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlErr, s, Config.data.vrb );
  HALT;  (* don't remove this before having checked return code implications *)
END Error;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Log.Write( mid, Log.lvlFat, s, TRUE );
  HALT;
END Fatal;


PROCEDURE MkPth( VAR p: ARRAY OF CHAR; r: UFTP.TResID );
BEGIN
  Str.Copy( p, r );
END MkPth;

PROCEDURE CalcMD5( f: FIO.File; VAR d: MD5.TDgstDesc ): LONGCARD;
CONST
  len = 4 * 1024;
VAR
  b: ARRAY [1..len] OF BYTE;
  l: CARDINAL;
  c: MD5.TCtxDesc;
  p: LONGCARD;
BEGIN
  FIO.Seek( f, 0 );
  p := 0;
  MD5.Init( c );
  l := FIO.RdBin( f, b, len );
  WHILE l > 0 DO
    INC( p, UFTP.TFileLen( l ) );
    MD5.Update( c, b, l );
    l := FIO.RdBin( f, b, len );
  END;
  MD5.Final( c, d );
  IF p # FIO.GetPos( f ) THEN (* happens after last successful read (l<len) *) END;
  RETURN p;
END CalcMD5;

PROCEDURE CalcMD5x( f: FIO.File; n: LONGCARD; VAR d: MD5.TDgstDesc );
CONST
  len = 4 * 1024;
VAR
  b: ARRAY [1..len] OF BYTE;
  l, m: CARDINAL;
  c: MD5.TCtxDesc;
BEGIN
  FIO.Seek( f, 0 );
  MD5.Init( c );
  WHILE n > 0 DO
    l := len;
    IF LONGCARD( l ) > n THEN l := CARDINAL( n ) END;
    m := FIO.RdBin( f, b, l );
    IF m # l THEN Fatal( 'CalcPos read error' ) END;
    MD5.Update( c, b, m );
    DEC( n, LONGCARD( m ) );
  END;
  MD5.Final( c, d );
END CalcMD5x;


PROCEDURE FCnt(): CARDINAL;
BEGIN
  RETURN fil.n;
END FCnt;

PROCEDURE Open( r: UFTP.TResID ): TFile;
VAR
  p: TPthStr;
  i: TVldIdx;
  f: TFile;
  b: FIO.File;
BEGIN
  MkPth( p, r );
  IF ~ FIO.Exists( p ) THEN RETURN NIL END;
  FOR i := MIN( TVldIdx ) TO MAX( TVldIdx ) DO
    f := ADR( fil.a[i] );
    IF f^.i = 0 THEN
      b := FIO.OpenRead( p );
      IF b = FIOStd.inv THEN RETURN NIL END;
      f^.i := i;
      f^.b := b;
      f^.r := r;
      f^.t.l := CalcMD5( b, f^.t.d );
      INC( fil.n );
      RETURN f;
    END;
  END;
  RETURN NIL;
END Open;

PROCEDURE GetRID( f: TFile; VAR r: UFTP.TResID );
BEGIN
  IF f = NIL THEN
    Lib2.Fill( r, 0 );
  ELSE
    r := f^.r;
  END;
END GetRID;

PROCEDURE GetTgt( f: TFile; VAR t: UFTP.TPosDesc );
BEGIN
  IF f = NIL THEN RETURN END;
  t := f^.t;
END GetTgt;

PROCEDURE GetPos( f: TFile; p: UFTP.TFileLen; VAR d: UFTP.TPosDesc );
BEGIN
  IF f = NIL THEN RETURN END;
  IF p >= f^.t.l THEN
    d := f^.t;
    RETURN;
  END;
  d.l := p;
  CalcMD5x( f^.b, p, d.d );
END GetPos;

PROCEDURE Read( f: TFile; p: LONGCARD; VAR x: ARRAY OF BYTE; l: CARDINAL ): CARDINAL;
VAR
  n: CARDINAL;
BEGIN
  IF ( f = NIL ) OR ( p > f^.t.l ) THEN RETURN 0 END;
  IF l > SIZE( x ) THEN l := SIZE( x ) END;
  FIO.Seek( f^.b, p );
  n := FIO.RdBin( f^.b, x, l );
  RETURN n;
END Read;

PROCEDURE Close( VAR f: TFile );
BEGIN
  IF f = NIL THEN RETURN END;
  FIO.Close( f^.b );
  Lib2.Fill( f^, 0 );
  f := NIL;
END Close;


PROCEDURE Cnt(): CARDINAL;
BEGIN
  RETURN snd.n;
END Cnt;

PROCEDURE Alloc( l: UFTP.TFgmLen; f: TFile ): TSndr;
VAR
  i: TVldIdx;
  s: TSndr;
BEGIN
  IF ( l = 0 ) OR ( f = NIL ) THEN RETURN NIL END;
  FOR i := MIN( TVldIdx ) TO MAX( TVldIdx ) DO
    s := ADR( snd.a[i] );
    IF s^.i = 0 THEN
      s^.i := i;
      s^.l := l;
      s^.f := f;
      s^.r.h := ( f^.t.l DIV l ) + ORD( f^.t.l MOD l > 0 );
      INC( snd.n );
      RETURN s;
    END;
  END;
  RETURN NIL;
END Alloc;

PROCEDURE GetFgl( s: TSndr; VAR l: UFTP.TFgmLen );
BEGIN
  IF s = NIL THEN l := 0 ELSE l := s^.l END;
END GetFgl;

PROCEDURE Receive( s: TSndr; i: UFTP.TFgmIdx; x: UFTP.TFgmSet );

  PROCEDURE Add( VAR r: TReqsDesc; i: UFTP.TFgmIdx );
  BEGIN
    IF ( r.n >= cntMaxReq ) OR ( i > r.h ) THEN RETURN END;
    r.a[ r.w ] := i;
    r.w := ( r.w + 1 ) MOD cntMaxReq;
    INC( r.n );
  END Add;

VAR
  j: UFTP.TRelFgmIdx;
BEGIN
  IF s = NIL THEN RETURN END;
  FOR j := MIN( UFTP.TRelFgmIdx ) TO MAX( UFTP.TRelFgmIdx ) DO
    IF j IN x THEN Add( s^.r, i + j ) END;
  END;
END Receive;

PROCEDURE Update( s: TSndr; n: CARDINAL; p: TSndProc; r: TRef );

  PROCEDURE Get( VAR r: TReqsDesc; VAR i: UFTP.TFgmIdx );
  BEGIN
    i := r.a[ r.r ];
    r.r := ( r.r + 1 ) MOD cntMaxReq;
    DEC( r.n );
  END Get;

VAR
  i: UFTP.TFgmIdx;
  l: CARDINAL;
  b: ARRAY [1..UFTP.lenMaxFgm] OF BYTE;
BEGIN
  IF s = NIL THEN RETURN END;
  WHILE ( n > 0 ) & ( s^.r.n > 0 ) DO
    Get( s^.r, i );
    l := s^.l;
    l := Read( s^.f, i * l, b, l );
    p( r, i, b, l );
    DEC( n );
  END;
END Update;

PROCEDURE Free( VAR s: TSndr );
BEGIN
  IF s = NIL THEN RETURN END;
  Lib2.Fill( s^, 0 );
  s := NIL;
END Free;


BEGIN
  Log.Register( mid, 'Sndrs' );
  Lib2.Fill( fil, 0 );
  Lib2.Fill( snd, 0 );
END Sndrs.
